@@ -3,6 +3,7 @@ from typing import (Iterable,
 
 from orient.hints import (Coordinate,
                           Point)
+from .utils import flatten
 
 BoundingBox = Tuple[Coordinate, Coordinate, Coordinate, Coordinate]
 
@@ -20,10 +21,14 @@ def disjoint_with(goal: BoundingBox, test: BoundingBox) -> bool:
             or goal_y_max < test_y_min or test_y_max < goal_y_min)
 
 
-def from_points(points: Iterable[Point]) -> BoundingBox:
+def from_iterable(points: Iterable[Point]) -> BoundingBox:
     iterator = iter(points)
     x_min, y_min = x_max, y_max = next(iterator)
     for x, y in iterator:
         x_min, x_max, y_min, y_max = (min(x_min, x), max(x_max, x),
                                       min(y_min, y), max(y_max, y))
     return x_min, x_max, y_min, y_max
+
+
+def from_iterables(iterables: Iterable[Iterable[Point]]) -> BoundingBox:
+    return from_iterable(flatten(iterables))
