@@ -1,41 +1,47 @@
 from orient.hints import (Contour,
                           Multiregion,
+                          Multisegment,
                           Point,
                           Polygon,
                           Region,
                           Segment)
-from .multiregion import (relate_contour as relate_contour_to_multiregion,
+from . import bounding_box
+from .multiregion import (_relate_contour as relate_contour_to_multiregion,
+                          _relate_multisegment
+                          as relate_multisegment_to_multiregion,
+                          _relate_region as _relate_region_to_multiregion,
                           relate_multiregion as relate_multiregions,
                           relate_region as relate_region_to_multiregion,
                           relate_segment as relate_segment_to_multiregion)
-from .region import (relate_contour as relate_contour_to_region,
+from .region import (_relate_contour as relate_contour_to_region,
+                     _relate_multisegment as relate_multisegment_to_region,
+                     _relate_region as relate_regions,
                      relate_point as relate_point_to_region,
-                     relate_region as relate_regions,
                      relate_segment as relate_segment_to_region)
 from .relation import Relation
 
 
 def relate_point(polygon: Polygon, point: Point) -> Relation:
     border, holes = polygon
-    relation_with_border = relate_point_to_region(border, point)
-    if relation_with_border is Relation.WITHIN:
+    relation_without_holes = relate_point_to_region(border, point)
+    if relation_without_holes is Relation.WITHIN:
         for hole in holes:
             relation_with_hole = relate_point_to_region(hole, point)
             if relation_with_hole is Relation.WITHIN:
                 return Relation.DISJOINT
             elif relation_with_hole is Relation.COMPONENT:
                 return Relation.COMPONENT
-    return relation_with_border
+    return relation_without_holes
 
 
 def relate_segment(polygon: Polygon, segment: Segment) -> Relation:
     border, holes = polygon
-    relation_with_border = relate_segment_to_region(border, segment)
-    if (holes and (relation_with_border is Relation.WITHIN
-                   or relation_with_border is Relation.ENCLOSED)):
+    relation_without_holes = relate_segment_to_region(border, segment)
+    if (holes and (relation_without_holes is Relation.WITHIN
+                   or relation_without_holes is Relation.ENCLOSED)):
         relation_with_holes = relate_segment_to_multiregion(holes, segment)
         if relation_with_holes is Relation.DISJOINT:
-            return relation_with_border
+            return relation_without_holes
         elif relation_with_holes is Relation.TOUCH:
             return Relation.ENCLOSED
         elif relation_with_holes is Relation.CROSS:
@@ -48,17 +54,56 @@ def relate_segment(polygon: Polygon, segment: Segment) -> Relation:
             # segment is within holes
             return Relation.DISJOINT
     else:
-        return relation_with_border
+        return relation_without_holes
+
+
+def relate_multisegment(polygon: Polygon,
+                        multisegment: Multisegment) -> Relation:
+    return (_relate_multisegment(polygon, multisegment,
+                                 bounding_box.from_iterables(multisegment))
+            if multisegment
+            else Relation.DISJOINT)
+
+
+def _relate_multisegment(polygon: Polygon,
+                         multisegment: Multisegment,
+                         multisegment_bounding_box: bounding_box.BoundingBox
+                         ) -> Relation:
+    border, holes = polygon
+    relation_without_holes = relate_multisegment_to_region(
+            border, multisegment, multisegment_bounding_box)
+    if (holes and (relation_without_holes is Relation.WITHIN
+                   or relation_without_holes is Relation.ENCLOSED)):
+        relation_with_holes = relate_multisegment_to_multiregion(
+                holes, multisegment, multisegment_bounding_box)
+        if relation_with_holes is Relation.DISJOINT:
+            return relation_without_holes
+        elif relation_with_holes is Relation.TOUCH:
+            return Relation.ENCLOSED
+        elif relation_with_holes is Relation.CROSS:
+            return Relation.CROSS
+        elif relation_with_holes is Relation.COMPONENT:
+            return Relation.COMPONENT
+        elif relation_with_holes is Relation.ENCLOSED:
+            return Relation.TOUCH
+        else:
+            # multisegment is within holes
+            return Relation.DISJOINT
+    else:
+        return relation_without_holes
 
 
 def relate_contour(polygon: Polygon, contour: Contour) -> Relation:
     border, holes = polygon
-    relation_with_border = relate_contour_to_region(border, contour)
-    if holes and (relation_with_border is Relation.ENCLOSED
-                  or relation_with_border is Relation.WITHIN):
-        relation_with_holes = relate_contour_to_multiregion(holes, contour)
+    contour_bounding_box = bounding_box.from_iterable(contour)
+    relation_without_holes = relate_contour_to_region(border, contour,
+                                                      contour_bounding_box)
+    if holes and (relation_without_holes is Relation.ENCLOSED
+                  or relation_without_holes is Relation.WITHIN):
+        relation_with_holes = relate_contour_to_multiregion(
+                holes, contour, contour_bounding_box)
         if relation_with_holes is Relation.DISJOINT:
-            return relation_with_border
+            return relation_without_holes
         elif relation_with_holes is Relation.TOUCH:
             return Relation.ENCLOSED
         elif (relation_with_holes is Relation.CROSS
@@ -70,12 +115,15 @@ def relate_contour(polygon: Polygon, contour: Contour) -> Relation:
             # contour is within holes
             return Relation.DISJOINT
     else:
-        return relation_with_border
+        return relation_without_holes
 
 
 def relate_region(polygon: Polygon, region: Region) -> Relation:
     border, holes = polygon
-    relation_with_border = relate_regions(border, region)
+    region_bounding_box = bounding_box.from_iterable(region)
+    relation_with_border = relate_regions(border, region,
+                                          bounding_box.from_iterable(border),
+                                          region_bounding_box)
     if relation_with_border in (Relation.DISJOINT,
                                 Relation.TOUCH,
                                 Relation.OVERLAP,
@@ -87,8 +135,9 @@ def relate_region(polygon: Polygon, region: Region) -> Relation:
         return (Relation.ENCLOSES
                 if holes
                 else relation_with_border)
-    elif holes:
-        relation_with_holes = relate_region_to_multiregion(holes, region)
+    else:
+        relation_with_holes = _relate_region_to_multiregion(
+                holes, region, region_bounding_box)
         if relation_with_holes is Relation.DISJOINT:
             return relation_with_border
         elif relation_with_holes is Relation.WITHIN:
@@ -104,8 +153,6 @@ def relate_region(polygon: Polygon, region: Region) -> Relation:
             return Relation.TOUCH
         else:
             return Relation.OVERLAP
-    else:
-        return relation_with_border
 
 
 def relate_multiregion(polygon: Polygon, multiregion: Multiregion) -> Relation:
@@ -145,55 +192,71 @@ def relate_multiregion(polygon: Polygon, multiregion: Multiregion) -> Relation:
 def relate_polygon(goal: Polygon, test: Polygon) -> Relation:
     goal_border, goal_holes = goal
     test_border, test_holes = test
-    borders_relation = relate_regions(goal_border, test_border)
-    if borders_relation in (Relation.DISJOINT, Relation.TOUCH,
+    return _relate_polygon(goal_border, goal_holes,
+                           bounding_box.from_iterable(goal_border),
+                           test_border, test_holes,
+                           bounding_box.from_iterable(test_border))
+
+
+def _relate_polygon(goal_border: Region,
+                    goal_holes: Multiregion,
+                    goal_border_bounding_box: bounding_box.BoundingBox,
+                    test_border: Region,
+                    test_holes: Multiregion,
+                    test_border_bounding_box: bounding_box.BoundingBox
+                    ) -> Relation:
+    borders_relation = relate_regions(goal_border, test_border,
+                                      goal_border_bounding_box,
+                                      test_border_bounding_box)
+    if borders_relation in (Relation.DISJOINT,
+                            Relation.TOUCH,
                             Relation.OVERLAP):
         return borders_relation
     elif borders_relation is Relation.EQUAL:
         if goal_holes and test_holes:
-            relation_with_holes = relate_multiregions(test_holes, goal_holes)
-            if relation_with_holes in (Relation.DISJOINT,
-                                       Relation.TOUCH,
-                                       Relation.OVERLAP):
+            holes_relation = relate_multiregions(test_holes, goal_holes)
+            if holes_relation in (Relation.DISJOINT,
+                                  Relation.TOUCH,
+                                  Relation.OVERLAP):
                 return Relation.OVERLAP
-            elif relation_with_holes is Relation.COVER:
+            elif holes_relation is Relation.COVER:
                 return Relation.ENCLOSES
-            elif relation_with_holes is Relation.WITHIN:
+            elif holes_relation is Relation.WITHIN:
                 return Relation.ENCLOSED
             else:
-                return relation_with_holes
+                return holes_relation
         else:
             return (Relation.ENCLOSES
                     if goal_holes
                     else (Relation.ENCLOSED
                           if test_holes
                           else Relation.EQUAL))
-    elif borders_relation in (Relation.WITHIN, Relation.ENCLOSED,
+    elif borders_relation in (Relation.WITHIN,
+                              Relation.ENCLOSED,
                               Relation.COMPONENT):
         if goal_holes:
-            relation_with_border = relate_region_to_multiregion(goal_holes,
-                                                                test_border)
-            if relation_with_border is Relation.DISJOINT:
+            holes_border_relation = _relate_region_to_multiregion(
+                    goal_holes, test_border, test_border_bounding_box)
+            if holes_border_relation is Relation.DISJOINT:
                 return borders_relation
-            elif relation_with_border is Relation.WITHIN:
+            elif holes_border_relation is Relation.WITHIN:
                 return Relation.DISJOINT
-            elif relation_with_border in (Relation.ENCLOSED,
-                                          Relation.COMPONENT,
-                                          Relation.EQUAL):
+            elif holes_border_relation in (Relation.ENCLOSED,
+                                           Relation.COMPONENT,
+                                           Relation.EQUAL):
                 return Relation.TOUCH
-            elif (relation_with_border is Relation.OVERLAP
-                  or relation_with_border is Relation.COMPOSITE):
+            elif (holes_border_relation is Relation.OVERLAP
+                  or holes_border_relation is Relation.COMPOSITE):
                 return Relation.OVERLAP
-            elif relation_with_border is Relation.TOUCH:
+            elif holes_border_relation is Relation.TOUCH:
                 return Relation.ENCLOSED
             elif test_holes:
-                relation_with_holes = relate_multiregions(test_holes,
-                                                          goal_holes)
+                holes_relation = relate_multiregions(test_holes, goal_holes)
                 return (borders_relation
-                        if relation_with_holes in (Relation.EQUAL,
-                                                   Relation.COMPONENT,
-                                                   Relation.ENCLOSED,
-                                                   Relation.WITHIN)
+                        if holes_relation in (Relation.EQUAL,
+                                              Relation.COMPONENT,
+                                              Relation.ENCLOSED,
+                                              Relation.WITHIN)
                         else Relation.OVERLAP)
             else:
                 return Relation.OVERLAP
@@ -201,35 +264,33 @@ def relate_polygon(goal: Polygon, test: Polygon) -> Relation:
             return (Relation.ENCLOSED
                     if test_holes and borders_relation is Relation.COMPONENT
                     else borders_relation)
-    else:
-        if test_holes:
-            relation_with_border = relate_region_to_multiregion(test_holes,
-                                                                goal_border)
-            if relation_with_border is Relation.DISJOINT:
-                return borders_relation
-            elif relation_with_border is Relation.WITHIN:
-                return Relation.DISJOINT
-            elif relation_with_border in (Relation.ENCLOSED,
+    elif test_holes:
+        holes_border_relation = _relate_region_to_multiregion(
+                test_holes, goal_border, goal_border_bounding_box)
+        if holes_border_relation is Relation.DISJOINT:
+            return borders_relation
+        elif holes_border_relation is Relation.WITHIN:
+            return Relation.DISJOINT
+        elif holes_border_relation in (Relation.ENCLOSED,
+                                       Relation.COMPONENT,
+                                       Relation.EQUAL):
+            return Relation.TOUCH
+        elif (holes_border_relation is Relation.OVERLAP
+              or holes_border_relation is Relation.COMPOSITE):
+            return Relation.OVERLAP
+        elif holes_border_relation is Relation.TOUCH:
+            return Relation.ENCLOSES
+        elif goal_holes:
+            holes_relation = relate_multiregions(goal_holes, test_holes)
+            return (borders_relation
+                    if holes_relation in (Relation.EQUAL,
                                           Relation.COMPONENT,
-                                          Relation.EQUAL):
-                return Relation.TOUCH
-            elif (relation_with_border is Relation.OVERLAP
-                  or relation_with_border is Relation.COMPOSITE):
-                return Relation.OVERLAP
-            elif relation_with_border is Relation.TOUCH:
-                return Relation.ENCLOSES
-            elif goal_holes:
-                relation_with_holes = relate_multiregions(goal_holes,
-                                                          test_holes)
-                return (borders_relation
-                        if relation_with_holes in (Relation.EQUAL,
-                                                   Relation.COMPONENT,
-                                                   Relation.ENCLOSED,
-                                                   Relation.WITHIN)
-                        else Relation.OVERLAP)
-            else:
-                return Relation.OVERLAP
+                                          Relation.ENCLOSED,
+                                          Relation.WITHIN)
+                    else Relation.OVERLAP)
         else:
-            return (Relation.ENCLOSES
-                    if goal_holes and borders_relation is Relation.COMPOSITE
-                    else borders_relation)
+            return Relation.OVERLAP
+    else:
+        return (Relation.ENCLOSES
+                if goal_holes and borders_relation is Relation.COMPOSITE
+                else borders_relation)
